require "Window"
require "GameLib"

local FLexFashion = {}

local nScanMaxItems = 150000
local nScanPerTick = 300
local fScanInterval = 0.05
local nItemsPerPage = 100
local nEmptySlotId = -1
local strRamp = "CRB_DyeRampSprites:sprDyeRamp_"
local tEmptyDyeInfo = { nId = 0, nRampIndex = 0, strName = "" }
local tEmptySlotInfo = { nItemId = nEmptySlotId, tDyeInfo = {tEmptyDyeInfo, tEmptyDyeInfo, tEmptyDyeInfo}, bVisible = true }

local aSlots = {
	GameLib.CodeEnumItemSlots.Weapon,
	GameLib.CodeEnumItemSlots.Head,
	GameLib.CodeEnumItemSlots.Shoulder,
	GameLib.CodeEnumItemSlots.Chest,
	GameLib.CodeEnumItemSlots.Hands,
	GameLib.CodeEnumItemSlots.Legs,
	GameLib.CodeEnumItemSlots.Feet,
}

local tSlotNameToId = {
	[Apollo.GetString("InventorySlot_WeaponPrimary")] = GameLib.CodeEnumItemSlots.Weapon,
	[Apollo.GetString("InventorySlot_Head")] = GameLib.CodeEnumItemSlots.Head,
	[Apollo.GetString("InventorySlot_Shoulder")] = GameLib.CodeEnumItemSlots.Shoulder,
	[Apollo.GetString("InventorySlot_Chest")] = GameLib.CodeEnumItemSlots.Chest,
	[Apollo.GetString("InventorySlot_Hands")] = GameLib.CodeEnumItemSlots.Hands,
	[Apollo.GetString("InventorySlot_Legs")] = GameLib.CodeEnumItemSlots.Legs,
	[Apollo.GetString("InventorySlot_Feet")] = GameLib.CodeEnumItemSlots.Feet,
}

local tCategoryIdToIdx = {
	[1] = 2,	-- Light Armor
	[2] = 3,	-- Medium Armor
	[3] = 4,	-- Heavy Armor
	[24] = 1,	-- Claws
	[8] = 2,	-- Greatsword
	[108] = 3,	-- Heavy Gun
	[16] = 4,	-- Pistols
	[22] = 5,	-- Psyblade
	[12] = 6,	-- Resonators
}

function FLexFashion:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self
	
	self.strSetSelectedName = "None"
	self.tSets = {
		None = {
			[GameLib.CodeEnumItemSlots.Head] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Shoulder] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Chest] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Hands] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Legs] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Feet] = tEmptySlotInfo,
			[GameLib.CodeEnumItemSlots.Weapon] = tEmptySlotInfo
		}
	}
	self.tSetDisplayed = TableUtil:Copy(self.tSets.None)
	
    return o
end

function FLexFashion:Init()
    Apollo.RegisterAddon(self)
end

function FLexFashion:OnLoad()
	Apollo.LoadSprites("Sprites.xml", "FLexFashionSprites");
	self.xmlDoc = XmlDoc.CreateFromFile("FLexFashion.xml")
	self.xmlDoc:RegisterCallback("OnDocumentReady", self)
end

function FLexFashion:OnDocumentReady()
	if not self.xmlDoc then
		return
	end
	
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuLoaded", self)
	
	Apollo.RegisterEventHandler("GenericEvent_OpenFLexFashion", "OnInit", self)
	
	self.tWndFilters = {}
	self.tWndSlots = {}
	self.tWndSelectedChannels = {}
	self.bIsSheathed = false
	
	self.unitPlayer = nil
end

function FLexFashion:OnInterfaceMenuLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "fLex Fashion", {"GenericEvent_OpenFLexFashion", "", "FLexFashionSprites:MaskIcon"})
end

function FLexFashion:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.General then
		return
	end

	local tSavedData = {
		tSets = self.tSets,
		strSetSelectedName = self.strSetSelectedName,
		tSetDisplayed = self.tSetDisplayed
	}

	return tSavedData
end

function FLexFashion:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.General then
		return
	end
	
	if tSavedData.tSets then
		self.tSets = tSavedData.tSets
	end
	
	if tSavedData.strSetSelectedName then
		self.strSetSelectedName = tSavedData.strSetSelectedName
	end
	
	if tSavedData.tSetDisplayed then
		self.tSetDisplayed = tSavedData.tSetDisplayed
	end
end

-----------
-- Setup --
-----------
function FLexFashion:OnInit()
	if self.wndMain then
		self:OnCloseConfirm()
	end
	
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "FLexFashionForm", nil, self)
	
	Event_FireGenericEvent("WindowManagementAdd", {wnd = self.wndMain, strName = "fLex Fashion"})
	
	self.wndModel = self.wndMain:FindChild("Model")
	self.wndBlocker = self.wndMain:FindChild("Blocker")
	self.wndProgress = self.wndBlocker:FindChild("Progress")
	self.tmrScanItems = ApolloTimer.Create(fScanInterval, true, "ScanItems", self)
	self.tmrScanItems:Stop()
	
	self.unitPlayer = GameLib.GetPlayerUnit()
	
	-- Filters --
	local wndFilters = self.wndMain:FindChild("Left:Filters")
	
	local aFilterGroups = { [2] = "Slots" }
	local aFilters = {
		{},
		{ Apollo.GetString("InventorySlot_Head"), Apollo.GetString("InventorySlot_Shoulder"), Apollo.GetString("InventorySlot_Chest"), Apollo.GetString("InventorySlot_Hands"), Apollo.GetString("InventorySlot_Legs"), Apollo.GetString("InventorySlot_Feet")},
		{}
	}
	
	for _, tFamily in ipairs(MarketplaceLib.GetAuctionableFamilies()) do
		if tFamily.nId == 26 then
			aFilters[1][1] = tFamily.strName
		elseif tFamily.nId == 1 or tFamily.nId == 2 then
			local nFamilyIdx
			
			if tFamily.nId == 1 then
				nFamilyIdx = 1
			elseif tFamily.nId == 2 then
				nFamilyIdx = 3
			end
			
			aFilterGroups[nFamilyIdx] = tFamily.strName
			
			for _, tCategory in pairs(MarketplaceLib.GetAuctionableCategories(tFamily.nId) or {}) do
				local nCategoryIdx = tCategoryIdToIdx[tCategory.nId]
					
				if tFamily.nId == 1 then
					aFilters[nFamilyIdx][nCategoryIdx] = tCategory.strName
				elseif tFamily.nId == 2 then
					for _, tType in pairs(MarketplaceLib.GetAuctionableTypes(tCategory.nId) or {}) do
						aFilters[nFamilyIdx][nCategoryIdx] = tType.strName 
					end
				end
			end
		end
	end
	
	wndFilters:FindChild("ReqLevel"):SetText(string.sub(Apollo.GetString("Tooltips_ReqLevel"), 0, -4))
	
	self.tWndFilters = {}
	self.tWndFilters.Name = wndFilters:FindChild("Name:Text")
	self.tWndFilters.Low = 0
	self.tWndFilters.High = 50
	
	local wndFilterList = wndFilters:FindChild("List")
	
	for nGroupId = 1, #aFilterGroups do
		local strGroup = aFilterGroups[nGroupId]
		local wndGroup = Apollo.LoadForm(self.xmlDoc, "FilterGroup", wndFilterList, self)
		local wndButtons = wndGroup:FindChild("Buttons")
		local btnAll = wndGroup:FindChild("All")
		
		btnAll:SetText(strGroup)
		btnAll:SetData(nGroupId)
		
		self.tWndFilters[nGroupId] = {}
		self.tWndFilters[nGroupId][0] = btnAll
		
		for nCategory = 1, #aFilters[nGroupId] do
			local strFilter = aFilters[nGroupId][nCategory]
			local btnFilter = Apollo.LoadForm(self.xmlDoc, "Filter", wndButtons, self)
			
			btnFilter:SetText(strFilter)
			btnFilter:SetData(nGroupId)
			
			if nGroupId == 2 then
				nCategoryId = tSlotNameToId[strFilter]
			else
				nCategoryId = nCategory
			end
			
			self.tWndFilters[nGroupId][nCategoryId] = btnFilter
		end
		
		wndButtons:ArrangeChildrenTiles()
	end
	
	wndFilterList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
	wndFilterList:RecalculateContentExtents()
	
	-- Items --
	local wndItemList = self.wndMain:FindChild("Left:Items:List")
	
	for idx = 1, nItemsPerPage do
		local wndItem = Apollo.LoadForm(self.xmlDoc, "Item", wndItemList, self)
		wndItem:SetName("Item" .. idx)
	end
	
	wndItemList:ArrangeChildrenVert()
	
	local wndItems = wndItemList:GetChildren()
	
	for _, wndItem in ipairs(wndItems) do
		wndItem:Show(false, true)
	end
	
	wndItemList:RecalculateContentExtents()
	
	-- Dyes --
	local wndDyeList = self.wndMain:FindChild("Right:Dyes:List")
	
	local tDyes = dofile(string.match(Apollo.GetAssetFolder(), "(.-)[\\/][Aa][Dd][Dd][Oo][Nn][Ss]") .. "\\Addons\\FLexFashion\\Dyes.lua")
	
	for i = 1, #tDyes do
		local tDye = tDyes[i]
		
		if tDye.nItemId then
			tDye.strName = Item.GetDataFromId(tDye.nItemId):GetName()
		end
		
		local wndDye = Apollo.LoadForm(self.xmlDoc, "Dye", wndDyeList, self)
		
		wndDye:FindChild("Swatch"):SetSprite(strRamp .. tDye.nRampIndex)
		wndDye:FindChild("Title"):SetText(tDye.strName)
		wndDye:FindChild("Button"):SetData(tDye)
	end
	
	wndDyeList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
	
	-- Sets --
	local aSlotItemArt = {
		"CharacterWindowSprites:btnCh_Armor_Hands",
		"CharacterWindowSprites:btnCh_Armor_Head",
		"CharacterWindowSprites:btnCh_Armor_Shoulder",
		"CharacterWindowSprites:btnCh_Armor_Chest",
		"CharacterWindowSprites:btnCh_Armor_Hands",
		"CharacterWindowSprites:btnCh_Armor_Legs",
		"CharacterWindowSprites:btnCh_Armor_Feet",
	}
	local wndSlots = self.wndMain:FindChild("Right:Sets:Slots")
	local wndColumns = Apollo.LoadForm(self.xmlDoc, "Columns", wndSlots, self)
	
	wndColumns:FindChild("Column1"):SetData(1)
	wndColumns:FindChild("Column2"):SetData(2)
	wndColumns:FindChild("Column3"):SetData(3)
	
	local wndSlotList = wndSlots:FindChild("List")
	
	for idx = 1, #aSlots do
		local eSlot = aSlots[idx]
		local wndSlot = Apollo.LoadForm(self.xmlDoc, "Slot", wndSlotList, self)
		
		wndSlot:SetData(eSlot)
		wndSlot:FindChild("Visible"):SetData(eSlot)
		wndSlot:FindChild("Item"):ChangeArt(aSlotItemArt[idx])
		
		self.tWndSlots[eSlot] = wndSlot
	end
	
	wndSlotList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
	
	local btnSetsDropDown = self.wndMain:FindChild("Right:Sets:DropDown")
	local wndSetsContainer = self.wndMain:FindChild("Right:Sets:Container")
	local wndSetsList = wndSetsContainer:FindChild("List")
	
	if self.strSetSelectedName ~= "None" then
		btnSetsDropDown:FindChild("Delete"):Show(true)
	else
		btnSetsDropDown:FindChild("Delete"):Show(false)
	end
	
	for strSetName, tSet in pairs(self.tSets) do
		local btnSet = Apollo.LoadForm(self.xmlDoc, "Set", wndSetsList, self)
		
		btnSet:SetText(strSetName)
		btnSet:SetData(strSetName)
		btnSet:SetName(strSetName)
		
		if self.strSetSelectedName == strSetName then
			btnSet:SetCheck(true)
			btnSetsDropDown:SetText(strSetName)
		end
	end
	
	wndSetsList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
	wndSetsList:RecalculateContentExtents()
	btnSetsDropDown:AttachWindow(wndSetsContainer)
	
	self.wndMain:FindChild("Model:Sheathed"):SetCheck(true)
	self:SetRedraw()
end

-------------
-- Filters --
-------------
function FLexFashion:OnFilterTextChanged(wndHandler, wndControl, strText)
	if strText and strText ~= "" then
		wndHandler:GetParent():FindChild("Clear"):Show(true)
	else
		wndHandler:GetParent():FindChild("Clear"):Show(false)
	end
end

function FLexFashion:OnFilterTextClear(wndHandler)
	wndHandler:GetParent():FindChild("Text"):SetText("")
	wndHandler:Show(false)
end

function FLexFashion:OnFilterLowChanged(wndHandler, wndControl, strText)
	local nLow = tonumber(strText)
	
	if nLow and nLow >= 1 and nLow <= 50 then
		self.tWndFilters.Low = nLow
	else
		wndControl:SetText(self.tFilters.Low)
	end
end

function FLexFashion:OnFilterHighChanged(wndHandler, wndControl, strText)
	local nHigh = tonumber(strText)
	
	if nHigh and nHigh >= 1 and nHigh <= 50 then
		self.tWndFilters.High = nHigh
	else
		wndControl:SetText(self.tFilters.High)
	end
end

function FLexFashion:OnFilterToggle(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	
	local nGroupId = wndHandler:GetData()
	
	for _, btnFilter in pairs(self.tWndFilters[nGroupId]) do
		if not btnFilter:IsChecked() then
			self.tWndFilters[nGroupId][0]:SetCheck(false)
			return
		end
	end
	
	self.tWndFilters[nGroupId][0]:SetCheck(true)
end

function FLexFashion:OnFilterGroupToggle(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	
	local nGroupId = wndHandler:GetData()
	
	for _, btnFilter in pairs(self.tWndFilters[nGroupId]) do
		btnFilter:SetCheck(wndHandler:IsChecked())
	end
end

-----------
-- Items --
-----------
function FLexFashion:ScanItems()
	local nNextIndex
	
	if self.nScanIndex <= nScanMaxItems then
		nNextIndex = math.min(self.nScanIndex + nScanPerTick, nScanMaxItems)
		
		for nItemId = self.nScanIndex, nNextIndex do
			local tItem = self:GetItemInfo(Item.GetDataFromId(nItemId))
			if tItem then
				table.insert(self.tItems, tItem)
			end
		end
		
		self.wndProgress:SetProgress(self.nScanIndex)
		self.nScanIndex = nNextIndex + 1
	else
		self.tmrScanItems:Stop()
		table.sort(self.tItems,
			function(a, b)
				return a.strName < b.strName
			end
		)
		self.nTotalPages = math.ceil(#self.tItems / nItemsPerPage)
		self:SetPage(1)
		self.wndBlocker:Show(false)
	end
end

function FLexFashion:GetItemInfo(item)
	if item == nil or item:IsDeprecated() then
		return
	end
	
	local nFamily = item:GetItemFamily()
	local nGroup
	
	if nFamily == 26 or nFamily == 1 then -- Costume or Armor
		nGroup = 1
	elseif nFamily == 2 then -- Weapon
		nGroup = 3
	end
	
	if not nGroup or not self.tWndFilters[nGroup] then
		return
	end
	
	local nCategory
	
	if nFamily == 26 then -- Costume
		nCategory = 1
	else
		nCategory = tCategoryIdToIdx[item:GetItemCategory()]
	end
	
	if not nCategory or not self.tWndFilters[nGroup][nCategory]:IsChecked() then
		return
	end
	
	local eSlot = tSlotNameToId[item:GetSlotName()]
	
	if not eSlot or not (eSlot == GameLib.CodeEnumItemSlots.Weapon or (self.tWndFilters[2][eSlot] and self.tWndFilters[2][eSlot]:IsChecked())) then
		return
	end
	
	local strName = string.lower(item:GetName())
	local strFilter = string.lower(self.tWndFilters.Name:GetText())
	
	if strFilter ~= "" and not string.find(strName, strFilter) then
		return
	end
	
	if strName == "" or string.match(strName, "_") then
		return
	end
	
	local nReqLevel = item:GetRequiredLevel()
	
	if not (nReqLevel >= self.tWndFilters.Low and nReqLevel <= self.tWndFilters.High) then
		return
	end
	
	return {["nItemId"] = item:GetItemId(), ["strName"] = item:GetName(), ["eSlot"] = eSlot}
end

function FLexFashion:Search()
	self.nScanIndex = 1
	self.tItems = {}
	self.wndProgress:SetFloor(0)
	self.wndProgress:SetMax(nScanMaxItems)
	self.wndProgress:SetProgress(0)
	self.wndBlocker:Show(true, true)
	self.tmrScanItems:Start()
end

function FLexFashion:SetPage(nPage)
	local wndItemList = self.wndMain:FindChild("Left:Items:List")
	
	for i = 1, nItemsPerPage do
		local tItem = self.tItems[(i + ((nPage - 1) * nItemsPerPage))]
		local wndItem = wndItemList:FindChild("Item" .. i)
		
		if tItem then
			local item = Item.GetDataFromId(tItem.nItemId)
			
			if not wndItem:IsShown() then
				wndItem:Show(true, true)
			end
			
			wndItem:SetData({nItemId = tItem.nItemId, eSlot = tItem.eSlot})
			wndItem:FindChild("Title"):SetText(tItem.strName)
			--wndItem:FindChild("Title"):SetTextColor(tQualityColors[tItem.nQuality])
			wndItem:FindChild("Icon"):GetWindowSubclass():SetItem(item)
		elseif wndItem:IsShown() then
			wndItem:Show(false, true)
		end
	end
	
	wndItemList:SetVScrollPos(0)
	wndItemList:RecalculateContentExtents()
	self.nPage = nPage
	self.wndMain:FindChild("Left:Items:Pages"):SetText(self.nPage .. " / " .. self.nTotalPages)
end

function FLexFashion:OnItemClick(wndHandler, wndControl, eMouseButton)
	if wndHandler ~= wndControl then
		return
	end
	
	if eMouseButton == GameLib.CodeEnumInputMouse.Right then
		local nItemId = wndHandler:GetData().nItemId
		Event_FireGenericEvent("GenericEvent_LinkItemToChat", Item.GetDataFromId(nItemId))
		wndControl:SetCheck(false)
	else
		local tData = wndHandler:GetData()
		
		if tData then
			self:SlotSetItem(tData.eSlot, tData.nItemId, {tEmptyDyeInfo, tEmptyDyeInfo, tEmptyDyeInfo}, true)
		end
	end
end

function FLexFashion:FirstPage()
	self:SetPage(1)
end

function FLexFashion:PreviousPage()
	if self.nPage > 1 then
		self:SetPage(self.nPage - 1)
	end
end

function FLexFashion:NextPage()
	if self.nPage < self.nTotalPages then
		self:SetPage(self.nPage + 1)
	end
end

function FLexFashion:LastPage()
	self:SetPage(self.nTotalPages)
end

----------
-- Dyes --
----------
function FLexFashion:OnDyeSearchChanged(wndHandler, wndControl, strText)
	local strSearch = string.lower(strText)
	
	local wndSearch = wndHandler:GetParent()
	local wndDyeList =  self.wndMain:FindChild("Right:Dyes:List")
	
	if strSearch and strSearch ~= "" then
		wndSearch:FindChild("Clear"):Show(true)
	else
		wndSearch:FindChild("Clear"):Show(false)
	end
	
	local tDyeList = wndDyeList:GetChildren()
	for idx, wndDye in pairs(tDyeList) do
		local strDye = string.lower(wndDye:FindChild("Button"):GetData().strName)
		if string.find(strDye, strSearch) then
			wndDye:Show(true, true)
		else
			wndDye:Show(false, true)
		end
	end
	
	wndDyeList:SetVScrollPos(0)
	wndDyeList:ArrangeChildrenVert()
	wndDyeList:RecalculateContentExtents()
end

function FLexFashion:OnDyeSearchClear(wndHandler, wndControl)
	local wndSearch = wndHandler:GetParent()
	local wndDyeList =  self.wndMain:FindChild("Right:Dyes:List")

	wndSearch:FindChild("Text"):SetText("")
	wndControl:Show(false)
	
	local tDyeList = wndDyeList:GetChildren()
	for idx, wndDye in pairs(tDyeList) do
		wndDye:Show(true, true)
	end
	
	wndDyeList:SetVScrollPos(0)
	wndDyeList:ArrangeChildrenVert()
	wndDyeList:RecalculateContentExtents()
end

function FLexFashion:OnDyeColorChecked(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	
	local tDyeInfo = wndHandler:GetData()
	
	for eSlot, tDyeWindows in pairs(self.tWndSelectedChannels) do
		local bSlotChanged = false
		
		for nChannel, wndDye in pairs(tDyeWindows) do
			self:UpdateSlotDye(eSlot, nChannel, tDyeInfo)
			
			bSlotChanged = true
		end
		
		if bSlotChanged then
			self.wndModel:SetItemDye(Item.GetDataFromId(self.tSetDisplayed[eSlot].nItemId), self.tSetDisplayed[eSlot].tDyeInfo[1].nId, self.tSetDisplayed[eSlot].tDyeInfo[2].nId, self.tSetDisplayed[eSlot].tDyeInfo[3].nId)
		end
	end
end

function FLexFashion:UpdateSlotDye(eSlot, nChannel, tDyeInfo)
	self.tSetDisplayed[eSlot].tDyeInfo[nChannel] = tDyeInfo
	
	local wndChannel = self.tWndSlots[eSlot]:FindChild("Dye" .. nChannel)
	wndChannel:FindChild("Hack:Swatch"):SetSprite(strRamp .. tDyeInfo.nRampIndex)
	wndChannel:SetData({eSlot = eSlot, nDyeChannel = nChannel, tDyeInfo = tDyeInfo})
	wndChannel:SetTooltip(tDyeInfo.strName)
end

function FLexFashion:OnDyeChannelToggle(wndHandler, wndControl)
	local tData = wndHandler:GetData()
	local eSlot = tData.eSlot
	
	if wndControl:IsChecked() then
		if not self.tWndSelectedChannels[eSlot] then
			self.tWndSelectedChannels[eSlot] = {}
		end
		
		self.tWndSelectedChannels[eSlot][tData.nChannel] = wndHandler:GetParent()
	elseif self.tWndSelectedChannels[eSlot] and self.tWndSelectedChannels[eSlot][tData.nChannel] then
		self.tWndSelectedChannels[eSlot][tData.nChannel] = nil
	end
end

function FLexFashion:OnDyeColumnToggle(wndHandler, wndControl)
	local nChannel = wndHandler:GetData()
	
	for idx = 1, #aSlots do
		local wndSlot = self.tWndSlots[aSlots[idx]]
		local wndDye = wndSlot:FindChild("Dye" .. nChannel .. ":Button")
		
		if wndSlot:IsShown() and wndDye:IsEnabled() then
			wndDye:SetCheck(wndControl:IsChecked())
			self:OnDyeChannelToggle(wndDye, wndDye)
		end
	end
end

----------
-- Sets --
----------
function FLexFashion:OnSetBtnChecked(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	
	self.strSetSelectedName = wndHandler:GetData()
	
	self:SwapSet()
end

function FLexFashion:OnSetDelete(wndHandler, wndControl)
	if self.strSetSelectedName ~= "None" then
		local wndSetList = self.wndMain:FindChild("Right:Sets:Container:List")
		
		wndSetList:FindChild(self.strSetSelectedName):Destroy()
		wndSetList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
		wndSetList:SetVScrollPos(0)
		wndSetList:RecalculateContentExtents()
		
		self.tSets[self.strSetSelectedName] = nil
		
		wndSetList:FindChild("None"):SetCheck(true)
		self.strSetSelectedName = "None"
		
		self:SwapSet()
	end
end

function FLexFashion:SwapSet()
	self.tSetDisplayed = TableUtil:Copy(self.tSets[self.strSetSelectedName])

	local tBtnSets = self.wndMain:FindChild("Right:Sets:Container:List"):GetChildren()
	
	if self.strSetSelectedName ~= "None" then
		self.wndMain:FindChild("Right:Sets:DropDown:Delete"):Show(true)
	else
		self.wndMain:FindChild("Right:Sets:DropDown:Delete"):Show(false)
	end
	
	self.wndMain:FindChild("Right:Sets:Container"):Show(false)
	self.wndMain:FindChild("Right:Sets:DropDown"):SetText(self.strSetSelectedName)
	
	self.tWndSelectedChannels = {}
	
	self:ColumnsResetControls()
	self:SetRedraw()
end

function FLexFashion:SetRedraw()
	if not self.wndMain then
		return
	end
	
	self.wndModel:SetCostume(self.unitPlayer)
	
	if self.tSetDisplayed then
		for idx = 1, #aSlots do
			local eSlot = aSlots[idx]
			
			if self.tSetDisplayed[eSlot].nItemId ~= -1 then
				self:SlotSetItem(eSlot, self.tSetDisplayed[eSlot].nItemId, self.tSetDisplayed[eSlot].tDyeInfo, self.tSetDisplayed[eSlot].bVisible)
			else
				self:SlotRemoveItem(eSlot)
			end
		end
	end
end

function FLexFashion:OnSetSave()
	self.wndMain:FindChild("Right:Sets:Dialog"):Show(true)
end

function FLexFashion:OnSetSaveConfirm(wndHandler, wndControl)
	local strSetName = wndHandler:GetParent():FindChild("Text"):GetText()
	
	if strSetName and strSetName ~= "" and strSetName ~= "None" then
		self.tSets[strSetName] = TableUtil:Copy(self.tSetDisplayed)
		self.strSetSelectedName = strSetName
		
		wndControl:GetParent():Show(false)
		
		local wndSetList = self.wndMain:FindChild("Right:Sets:Container:List")
		local wndSets = wndSetList:GetChildren()
		
		for idx, wndSet in ipairs(wndSets) do
			wndSet:SetCheck(false)
		end
		
		local btnSet = Apollo.LoadForm(self.xmlDoc, "Set", wndSetList, self)
		
		btnSet:SetText(strSetName)
		btnSet:SetData(strSetName)
		btnSet:SetName(strSetName)
		btnSet:SetCheck(true)
		
		self.wndMain:FindChild("Right:Sets:DropDown"):SetText(strSetName)
		
		wndSetList:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.LeftOrTop)
		wndSetList:SetVScrollPos(0)
		wndSetList:RecalculateContentExtents()
	end
end

function FLexFashion:OnSetSaveCancel(wndHandler, wndControl)
	wndHandler:GetParent():FindChild("Text"):SetText("")
	wndControl:GetParent():Show(false)
end

function FLexFashion:OnSetReset()
	self:SwapSet()
end

function FLexFashion:SlotSetItem(eSlot, nItemId, tDyeInfo, bVisible)
	local item = Item.GetDataFromId(nItemId)
	
	if item then
		local wndSlot = self.tWndSlots[eSlot]
		local wndSlotItem = wndSlot:FindChild("Item")
		local wndSlotIcon = wndSlotItem:FindChild("Icon")
		local tAvailableChannels = item:GetAvailableDyeChannel()
		
		self.tSetDisplayed[eSlot].nItemId = nItemId
		wndSlotItem:SetData(nItemId)
		wndSlotIcon:SetTooltipForm(nil)
		wndSlotIcon:SetTooltipDoc(nil)
		wndSlotIcon:GetWindowSubclass():SetItem(item)
		
		for idx = 1, 3 do
			local wndDye = wndSlot:FindChild("Dye" .. idx)
			local btnButton = wndDye:FindChild("Button") -- Who's got the button?
			
			wndDye:FindChild("Hack:Swatch"):SetSprite(strRamp .. tDyeInfo[idx].nRampIndex)
			wndDye:FindChild("Hack:Swatch"):SetTooltip(tDyeInfo[idx].strName)
			btnButton:SetData({eSlot = eSlot, nChannel = idx, tDyeInfo = tDyeInfo[idx]})
			self.tSetDisplayed[eSlot].tDyeInfo[idx] = tDyeInfo[idx]
			
			btnButton:Enable(bVisible and (tAvailableChannels["bDyeChannel" .. idx] or false))
			btnButton:SetCheck(false)
		end
		
		if bVisible then
			self.wndModel:SetItem(item)
			self.wndModel:SetItemDye(item, tDyeInfo[1].nId, tDyeInfo[2].nId, tDyeInfo[3].nId)
		else
			self.wndModel:RemoveItem(eSlot)
		end
		
		wndSlot:FindChild("Visible"):Enable(true)
		wndSlot:FindChild("Visible"):SetCheck(bVisible)
		self:SlotClearSelectedChannels(eSlot)
	end
end

function FLexFashion:OnModelSheatheToggle(wndHandler, wndControl)
	self.wndModel:SetSheathed(wndControl:IsChecked())
end

function FLexFashion:SlotRemoveItem(eSlot)
	local wndSlot = self.tWndSlots[eSlot]
	
	self.tSetDisplayed[eSlot].nItemId = nEmptySlotId
	wndSlot:FindChild("Item:Icon"):GetWindowSubclass():SetItem(nil)
	
	for idx = 1, 3 do
		local wndDye = wndSlot:FindChild("Dye" .. idx)
		local btnButton = wndDye:FindChild("Button") -- Who's got the button?
		
		wndDye:FindChild("Hack:Swatch"):SetSprite("")
		wndDye:FindChild("Hack:Swatch"):SetTooltip("")
		btnButton:SetData({eSlot = eSlot, nChannel = idx, tDyeInfo = tEmptyDyeInfo,})
		self.tSetDisplayed[eSlot].tDyeInfo[idx] = tEmptyDyeInfo
		
		btnButton:Enable(false)
		btnButton:SetCheck(false)
	end
	
	self.wndModel:RemoveItem(eSlot)
	
	wndSlot:FindChild("Visible"):Enable(false)
	self:SlotClearSelectedChannels(eSlot)
end

function FLexFashion:OnSlotItemClick(wndHandler, wndControl, eMouseButton)
	if eMouseButton == GameLib.CodeEnumInputMouse.Right then
		self:SlotRemoveItem(wndHandler:GetParent():GetData())
	else
		local nItemId = wndHandler:GetData()
		Event_FireGenericEvent("GenericEvent_LinkItemToChat", Item.GetDataFromId(nItemId))
	end
end

function FLexFashion:OnSlotItemGenerateTooltip(wndHandler, wndControl, eType, itemCurr, idx)
	if wndHandler ~= wndControl then
		return
	end

	local itemData = wndHandler:GetData()
	local strAppend = "<P>" .. Apollo.GetString("Costumes_RemoveTooltip") .. "</P>"

	
	if itemData and Tooltip and Tooltip.GetItemTooltipForm then
		Tooltip.GetItemTooltipForm(self, wndHandler, itemData, {bSimple = true, strAppend = strAppend})
	end
end

function FLexFashion:OnSlotVisibleBtnToggle(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return false
	end

	local bVisible = wndHandler:IsChecked()
	local eSlot = wndHandler:GetData()
	local wndSlot = self.tWndSlots[eSlot]
	
	self.tSetDisplayed[eSlot].bVisible = bVisible
	
	if not bVisible then
		self.wndModel:RemoveItem(eSlot)
		self:SlotClearSelectedChannels(eSlot)
		wndSlot:FindChild("Dye1:Button"):Enable(false)
		wndSlot:FindChild("Dye2:Button"):Enable(false)
		wndSlot:FindChild("Dye3:Button"):Enable(false)
	else
		local item = Item.GetDataFromId(self.tSetDisplayed[eSlot].nItemId)
		local tDyeInfo = self.tSetDisplayed[eSlot].tDyeInfo
		self.wndModel:SetItem(item)
		self.wndModel:SetItemDye(item, tDyeInfo[1].nId, tDyeInfo[2].nId, tDyeInfo[3].nId)
		
		local tAvailableChannels = item:GetAvailableDyeChannel()
		for idx = 1, 3 do
			wndButton = wndSlot:FindChild("Dye" .. idx .. ":Button")
			wndButton:Enable((tAvailableChannels["bDyeChannel" .. idx] or false))
			if wndButton:IsEnabled() and self.wndMain:FindChild("Right:Sets:Slots:Columns:Column" .. idx):IsChecked() then
				wndButton:SetCheck(true)
			end
		end
	end
	
	wndSlot:FindChild("Dye1"):SetBGColor(bVisible and "UI_AlphaPercent100" or "UI_AlphaPercent50")
	wndSlot:FindChild("Dye2"):SetBGColor(bVisible and "UI_AlphaPercent100" or "UI_AlphaPercent50")
	wndSlot:FindChild("Dye3"):SetBGColor(bVisible and "UI_AlphaPercent100" or "UI_AlphaPercent50")
end

function FLexFashion:SlotClearSelectedChannels(eSlot)
	if self.tWndSelectedChannels[eSlot] then
		for nChannel, _  in pairs(self.tWndSelectedChannels[eSlot]) do
			self.tWndSelectedChannels[eSlot][nChannel]:FindChild("Button"):SetCheck(false)
		end
		
		self.tWndSelectedChannels[eSlot] = nil
	end
end

function FLexFashion:ColumnsResetControls()
	local wndColumns = self.wndMain:FindChild("Right:Sets:Slots:Columns")
	
	wndColumns:FindChild("Column1"):SetCheck(false)
	wndColumns:FindChild("Column2"):SetCheck(false)
	wndColumns:FindChild("Column3"):SetCheck(false)
end

function FLexFashion:OnCloseConfirm()
	self.tWndFilters = {}
	self.tWndSlots = {}
	self.tWndSelectedChannels = {}
	self.unitPlayer = nil
	
	self.wndMain:Close()
	self.wndMain:Destroy()
	self.wndMain = nil
	self.wndModel = nil
end

function FLexFashion:DyeSearchChanged(wndHandler, wndControl, strText)
	local strFilter = string.lower(wndControl:GetText())
	
	if strFilter and strFilter ~= "" then
		self.wndDyes:FindChild("Search:Clear"):Show(true)
	else
		self.wndDyes:FindChild("Search:Clear"):Show(false)
	end
	
	local tDyeList = self.wndDyeList:GetChildren()
	for idx, wndDye in pairs(tDyeList) do
		local strDye = string.lower(wndDye:FindChild("Hack:Title"):GetText())
		if string.find(strDye, strFilter) then
			wndDye:Show(true, true)
		else
			wndDye:Show(false, true)
		end
	end
	
	self.wndDyeList:SetVScrollPos(0)
	self.wndDyeList:ArrangeChildrenVert()
	self.wndDyeList:RecalculateContentExtents()
end

function FLexFashion:OnCategoryToggle(wndHandler, wndControl)
	for nCategoryId, wndCategory in pairs(self.wndMain:FindChild("Left:Filters:List"):GetChildren()) do
		local nListHeight = wndCategory:FindChild("Button"):IsChecked() and (wndCategory:FindChild("List"):ArrangeChildrenVert(0) + 15) or 0
		local nLeft, nTop, nRight, nBottom = wndCategory:GetAnchorOffsets()
		wndCategory:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nListHeight + 42)
	end

	self.wndMain:FindChild("Left:Filters:List"):ArrangeChildrenVert(0)
end

function FLexFashion:GetNewDyes()
	self.nScanIndex = 1
	self.tAllDyes = {}
	self.tmrScanDyes = ApolloTimer.Create(fScanInterval, true, "ScanDyes", self)
end

function FLexFashion:ScanDyes()
	local nNextIndex
	
	if self.nScanIndex <= nScanMaxItems then
		nNextIndex = math.min(self.nScanIndex + nScanPerTick, nScanMaxItems)
		
		for nItemId = self.nScanIndex, nNextIndex do
			local strName = self:GetDyeInfo(Item.GetDataFromId(nItemId))
			if strName then
				if not self.tAllDyes[strName] then
					self.tAllDyes[strName] = {}
				end
				
				table.insert(self.tAllDyes[strName], nItemId)
			end
		end
		
		self.nScanIndex = nNextIndex + 1
	else
		self.tmrScanDyes:Stop()
		
		local tDyes = dofile(string.match(Apollo.GetAssetFolder(), "(.-)[\\/][Aa][Dd][Dd][Oo][Nn][Ss]") .. "\\Addons\\FLexFashion\\Dyes.lua")
	
		for idx, tDye in ipairs(tDyes) do
			if self.tAllDyes[tDye.strName] then
				self.tAllDyes[tDye.strName] = nil
			end
		end
		
		ChatSystemLib.PostOnChannel(1, "[FLexFashion] Dye Scanner finished!")
	end
end

function FLexFashion:GetDyeInfo(item)
	if item == nil or item:IsDeprecated() then
		return
	end
	
	if item:GetItemFamily() ~= 16 then
		return
	end
	
	if item:GetItemCategory() ~= 54 then
		return
	end
	
	if not string.match(item:GetName(), "Dye") then
		return
	end
	
	return item:GetName()
end

local FLexFashionInst = FLexFashion:new()
FLexFashionInst:Init()